#define PAM_SM_AUTH
#include <stdio.h>
#include <unistd.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "sslauth.h"

static int check_console(pam_handle_t *pamh)
{
    int len;
    char *uttyname;
    char ttyfileline[256];
    struct stat ttyfileinfo;
    FILE *ttyfile;

    if ((pam_get_item(pamh, PAM_TTY, (const void **)&uttyname) != PAM_SUCCESS) || (uttyname == NULL) || (*uttyname==0)) {
        _pam_log(LOG_WARNING, "Cannot determine user's tty!");
	return PAM_SERVICE_ERR;
    }

	// The PAM_TTY item may be prefixed with "/dev/" - skip that
    if (strncmp(TTY_PREFIX, uttyname, sizeof(TTY_PREFIX)-1) == 0)
	uttyname += sizeof(TTY_PREFIX)-1;

	// If securetty file isn't pressent - return OK 
    if (stat(SECURETTY_FILE, &ttyfileinfo)) {
	_pam_log(LOG_NOTICE, "Couldn't open %s!", SECURETTY_FILE);
	return PAM_SUCCESS; 
    }

	// If the file is world writable or is not a normal file, return error
    if ((ttyfileinfo.st_mode & S_IWOTH)||!S_ISREG(ttyfileinfo.st_mode)) {
	_pam_log(LOG_ERR, "%s is either world writable or not a normal file!", SECURETTY_FILE);
	return PAM_AUTH_ERR;
    }

    ttyfile = fopen(SECURETTY_FILE,"r");
    if(ttyfile == NULL) {
	_pam_log(LOG_ERR,"Error opening %s! ", SECURETTY_FILE);
	return PAM_SERVICE_ERR;
    }

    while (fgets(ttyfileline,sizeof(ttyfileline)-1, ttyfile) != NULL) {
	len=strlen(ttyfileline)-1;
	while ((len>=0)&&((ttyfileline[len]=='\n')||(ttyfileline[len]=='\t')||(ttyfileline[len]=='\r')||(ttyfileline[len]==' '))) len--;
	ttyfileline[len+1]=0;
	if (!strcmp(ttyfileline,uttyname)) {
	    fclose(ttyfile);
	    return PAM_SUCCESS;	    
	}
    }
    fclose(ttyfile);
    _pam_log(LOG_NOTICE, "Access denied: tty '%s' is not secure!", uttyname);    
    return PAM_AUTH_ERR;
}

static int check_remote(pam_handle_t *pamh) {
    char *rhost;
    if (pam_get_item(pamh, PAM_RHOST, (const void **)&rhost)!=PAM_SUCCESS) {
        _pam_log(LOG_WARNING, "Cannot determine user's remote host!");
	return PAM_SERVICE_ERR;
    }
    if (!rhost) return PAM_SUCCESS;
    _pam_log(LOG_NOTICE, "Non local access denied!");
    return PAM_AUTH_ERR;
}

PAM_EXTERN int pam_sm_authenticate(pam_handle_t *pamh,int flags,int argc, const char **argv)
{
    int i;
    int retval;
    const char *user=NULL;			/* User Name */
    char *ruser;
    const char *ca_dir=NULL, *ca_file=NULL, *crl_dir=NULL, *crl_file=NULL, *key_dir=NULL;
    unsigned char bluetooth=0;
    char *btdir;
    
	// Setting configuration
    for (i=0;i<argc;i++) {
	if (strncmp(argv[i],"ca_dir=",7)==0) {
	    ca_dir=argv[i]+7;
	} else if (strncmp(argv[i],"ca_file=",8)==0) {
	    ca_file=argv[i]+8;
	} else if (strncmp(argv[i],"crl_dir=",8)==0) {
	    crl_dir=argv[i]+8;
	} else if (strncmp(argv[i],"crl_file=",9)==0) {
	    crl_file=argv[i]+9;
	} else if (strncmp(argv[i],"key_dir=",8)==0) {
	    key_dir=argv[i]+8;
#ifdef _BLUETOOTH
	} else if (strncmp(argv[i],"bluetooth",9)==0) {
	    bluetooth=1;
#endif
	} else {
	    _pam_log(LOG_WARNING, "Unknown parameter: %s",argv[i]);
	}
    }
    
	// Checking for supported console
    retval = check_console(pamh);
    if (retval != PAM_SUCCESS) return retval;

	// Check for local requestor
    retval = check_remote(pamh);
    if (retval != PAM_SUCCESS) return retval;
    
    if ((!ca_file)&&(!ca_dir)) {
	ca_file=(char*)alloca(strlen(default_ca)+1);
	if (!ca_file) {
	    _pam_log(LOG_ERR, "Can't allocate memory!");
	     return PAM_AUTH_ERR;
	}
	strcpy((char*)ca_file,default_ca);
    }

    if ((!crl_dir)&&(!crl_dir)) {
	crl_dir=(char*)alloca(strlen(default_crl)+1);
	if (!crl_dir) {
	    _pam_log(LOG_ERR, "Can't allocate memory!");
	     return PAM_AUTH_ERR;
	}
	strcpy((char*)crl_dir,default_crl);
    }
    
    if (!key_dir) {
	key_dir=(char*)alloca(strlen(default_keys)+1);
	if (!key_dir) {
	    _pam_log(LOG_ERR, "Can't allocate memory!");
	     return PAM_AUTH_ERR;
	}
	strcpy((char*)key_dir,default_keys);
    }

    retval = pam_get_user(pamh, &user, NULL);
    if (retval != PAM_SUCCESS) {
	_pam_log(LOG_ERR,"pam_get_user returned error: %s", pam_strerror(pamh,retval));
	return retval;
    }
    
    ruser=authenticate(user,pamh,ca_dir,ca_file,crl_dir,crl_file,key_dir);
    if (ruser) {
	if (user == NULL || *user == '\0') {
	    pam_set_item(pamh, PAM_USER, (const void *) ruser);
	}
	_pam_log(LOG_NOTICE,"User(%s) authenticated!", ruser);
	free(ruser);

	user = NULL;                                        /* clean up */
	return PAM_SUCCESS;
    } else {
#ifdef _BLUETOOTH
	if (bluetooth) btdir=BluetoothInit();
	else btdir=NULL;
	if (btdir) {
    	    ruser=authenticate(user,pamh,ca_dir,ca_file,crl_dir,crl_file,btdir);
	    BluetoothFree(btdir);
	}
	if (ruser) {
	    if (user == NULL || *user == '\0') {
		pam_set_item(pamh, PAM_USER, (const void *) ruser);
	    }
	    _pam_log(LOG_NOTICE,"User(%s) authenticated!", ruser);
	    free(ruser);

	    user = NULL;                                        /* clean up */
	    return PAM_SUCCESS;
	}
	else {
#endif
	    if ((user)&&(*user)) {
		_pam_log(LOG_NOTICE,"User(%s) authentication failed!", user);
	    } else {
		_pam_log(LOG_NOTICE,"Auto authentication failed!", user);
	    }
#ifdef _BLUETOOTH
	}
#endif
    }

    user = NULL;                                            /* clean up */
    return PAM_AUTH_ERR;
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t *pamh,int flags,int argc, const char **argv)
{
     return PAM_SUCCESS;
 
}


/* static module data */
#ifdef PAM_STATIC
struct pam_module _pam_deny_modstruct = {
    "pam_x509",
    pam_sm_authenticate,
    pam_sm_setcred,
    NULL,
    NULL,
    NULL,
    NULL
};
#endif
